# CBpanel Systray plugin

name = "Systray"
description = "System tray implementation using Xlib"

import CBpanel, gtk, gobject

import sys
from gtk import gdk
import cairo        # Awn and cairo drawing
from Xlib import X, display, error, Xatom, Xutil
import Xlib.protocol.event
                    # Xlib and bits that are needed.
                    # These are used to create the custom GTK widget
                    # and to interface with system tray icons
import subprocess
                    # used to launch the config

import atexit
                    # Used in an attempt at clean closing

class Plugin(CBpanel.Plugin):
    def __init__(self, panel, settings):
        CBpanel.Plugin.__init__(self, panel, settings)
        self.set_name("CBpanelStalonetray")
        self.astate = 0
        self.set_border_width(1)

        self.systray = mywidget(display, error, self, panel)
        self.add(self.systray)

    def destroy(self):
        self.logINFO("attempt to cleanly close, in such a way that the icons do not get an X window error")
        self.systray.cleanup()

class Obj(object):
#----------------------------------------------------------------------------
    """ Multi-purpose class """
    #----------------------------

    def __init__(self, **kwargs):
    #----------------------------
        self.__dict__.update(kwargs)

class mywidget(gtk.Widget):

    def __init__(self, display, error, gtkwin, panel):
        gtk.Widget.__init__(self)

        self.panel = panel

        self.BG_COLOR="0x0070E0"
        self.CUSTOM_Y=5
        self.HIGH=1
        self.ALLOW_COL=50
        self.DIVIDEBYZERO=False
        self.BORDER=False
        self.ZEROPID=True
        self.USEIM = False
        self.USEGTK = 1
        self.ICONSIZE= 24 # panel.iconsize

        # Define widget value and set to default
        self.curr_x=1 #Starting width and height
        self.curr_y=1

        self.dsp = display.Display()      # references to Xlib
        self.scr = self.dsp.screen()
        self.root = self.scr.root
        self.error = error.CatchError()

        self.gtkwin=gtkwin               # Applet's reference

        self.realized= 0                 # Is the Xwindow realized yet?
                                         # if not, can cause problems with
                                         # certain functions
        ourmask = (X.ButtonPressMask|X.ButtonReleaseMask|X.ExposureMask)

        self.wind = self.root.create_window(0, 0, 1, 10,
                0, self.scr.root_depth, window_class=X.InputOutput,
                visual=X.CopyFromParent, colormap=X.CopyFromParent,
                event_mask=ourmask)

        # System Tray window


        self.tray = Obj(id="tray", tasks={}, order=[],
                    first=0, last=0, window=self.wind)
        # Create an empty Object, this will contain all the data on icons
        # to be added, and all currently managed


        # Create a non-visible window to be the selection owner
        self._OPCODE = self.dsp.intern_atom("_NET_SYSTEM_TRAY_OPCODE")
        self.manager = self.dsp.intern_atom("MANAGER")
        self.selection = self.dsp.intern_atom(
            "_NET_SYSTEM_TRAY_S%d" % self.dsp.get_default_screen())
        self.selowin = self.scr.root.create_window(-1,
                                  -1, 1, 1, 0, self.scr.root_depth)
        owner = self.dsp.get_selection_owner(self.selection)
        if(owner!=X.NONE):
            # If someone already has the system tray... BAIL!
            self.gtkwin.logDEBUG("Another System Tray is already running")

        else:
            self.selowin.set_selection_owner(self.selection, X.CurrentTime)
            self.tr__sendEvent(self.root, self.manager,
                  [X.CurrentTime, self.selection, self.selowin.id],
                  (X.StructureNotifyMask))

            self.tr__setProps(self.dsp, self.wind)
            # Set a list of Properties that we'll need

            self.dsp.flush()
            # Show the window and flush the display

    def do_realize(self):
        self.set_flags(gtk.REALIZED)

        self.window=gdk.window_foreign_new(self.wind.id)
        self.window.reparent(self.gtkwin.window, 0, 0)
        # Take the system manager window (not selection owner!)
        # And make it the gdk.window of the custom widget.

        self.window.set_user_data(self)
        self.style.attach(self.window)
        # Set it up as a custom widget and tell it what style (theme) to use

        self.style.set_background(self.window, gtk.STATE_NORMAL)
        self.window.move_resize(*self.allocation)
        # Tell it to use the background colour as background colour...
        # Im sure theres a reason i need to tell it that ;)

        self.tr__updatePanel(self.root, self.wind)
        # First render. Grab all the icons we know about, tell them where to
        # draw, and call a resize if necessary (likely, the first time around)

        gobject.io_add_watch(self.dsp.fileno(), gobject.IO_IN | gobject.IO_PRI,
                             self.tr__testTiming)

        self.redraw()

        self.realized= 1
        # and now we can safely render alpha :D

    def do_unrealize(self):
        # The do_unrealized method is responsible for freeing the GDK resources
        return 1

    def do_size_request(self, requisition):
        # Widget is bieng asked what size it would like to be.
        requisition.width=self.curr_x
        requisition.height=self.curr_y

    def do_size_allocate(self, allocation):
        # The do_size_allocate is called by when the actual size is known
        # and the widget is told how much space could actually be allocated

        self.allocation = allocation

        # If we're realized, move and resize the window to the
        # requested coordinates/positions
        if self.flags() & gtk.REALIZED:
            self.window.move_resize(*allocation)

    def tr__taskDelete(self, tid):
    #--------------------------------
        """ Delete the given task ID if it's in the tray/task list """
        if tid in self.tray.tasks:
            del self.tray.tasks[tid]
            self.tray.order.remove(tid)
            return 1
        return 0

    def tr__updatePanel(self, root, win):
        # Requested re-draw/re-position
        rr= self.window.get_geometry()
        # find the gdk windows geometry (for the size of y)

        offsety=rr[3]-((self.HIGH*self.ICONSIZE)+self.CUSTOM_Y)
        if(offsety<0):
            offsety=0
        # find the Y position to start drawing icons at


        # First - a pre-render loop to find how much space we need.
        space=2 # Border of 1 pixel either side
        if(self.BORDER==True):
            space+=5 # For rounder corners, more border space is needed
        tempy=0
        for t in self.tray.tasks.values():
            iwant=0
            ifail=0
            try:
                iwant=t.obj.get_wm_normal_hints().min_width
            except:
                ifail=1
                pass
            if ifail==0:
                t.width = self.ICONSIZE
                #space += t.width
                if(tempy==0):
                    space+=t.width
                if(tempy < self.HIGH-1):
                    tempy+=1
                else:
                    tempy=0

        if(self.BORDER==True):
            space+=5
        self.set_size_request(space, self.CUSTOM_Y+self.HIGH*self.ICONSIZE)
        self.gtkwin.set_size_request(space, self.CUSTOM_Y+self.HIGH*self.ICONSIZE)
        # Request resize to the new size we need :)

        #Second pass, telling each icon where it is to go now.
        self.curr_x=0
        if(self.BORDER==True):
            self.curr_x+=5
        self.curr_y=0
        for tid in self.tray.order:
            t = self.tray.tasks[tid]
            t.x = self.curr_x
            t.y = offsety+self.curr_y*self.ICONSIZE
            t.obj.configure(onerror=self.error, x=t.x, y=t.y,
                            width=self.ICONSIZE, height=self.ICONSIZE)
            t.obj.map(onerror=self.error)
            if(self.curr_y < self.HIGH-1):
                self.curr_y+=1
            else:
                self.curr_x+=t.width
                self.curr_y=0
        if(self.curr_y == 0):
            self.curr_x+=1
        else:
            self.curr_x+=25
        self.curr_y=offsety+(self.HIGH)*self.ICONSIZE
        self.tr__updateAlpha(False)
        # And then update the alpha again, just to make sure

    def tr__updateAlpha(self, returnvar):
        rr= self.window.get_geometry()
        offsety=rr[3]-((self.HIGH*self.ICONSIZE)+self.CUSTOM_Y)
        # Again : find location of icons on the widget

        if(self.realized == 1 and offsety>-1):
            w= rr[2]
            h= offsety+ (self.HIGH*self.ICONSIZE)
            if(self.BORDER==True):
                h+=10

            # get the width and height again

            # Create a 1Bit-map, each pixel is either True of False
            # if i use a function to draw on certain pixels, those
            # that are set to True will be shown, all False will be
            # 100% transparent.
            if(self.DIVIDEBYZERO == False):
                bitmap = gtk.gdk.Pixmap(None, w, h, 1)
                cr = bitmap.cairo_create()

                # Clear the bitmap to False
                cr.set_source_rgb(0, 0, 0)
                cr.set_operator(cairo.OPERATOR_DEST_OUT)
                cr.paint()

                # Draw our shape into the bitmap using cairo
                cr.set_operator(cairo.OPERATOR_OVER)

                # Lets not do the impossible. Just draw a box around it.
                if(self.BORDER == True):
                    newh= (self.HIGH*self.ICONSIZE)
                    cr.set_line_width(1)

                    # For the rounded edges, I just hacked together some
                    # hardwired numbers that looked acceptable.
                    cr.rectangle(0, offsety+1, w-4, newh-2)
                    cr.fill()
                    cr.rectangle(6, offsety-5, w-16, newh+10)
                    cr.fill()
                    cr.rectangle(1, offsety-1, w-6, newh+2)
                    cr.fill()
                    cr.rectangle(2, offsety-2, w-8, newh+4)
                    cr.fill()
                    cr.rectangle(3, offsety-3, w-10, newh+6)
                    cr.fill()
                    cr.rectangle(4, offsety-4, w-12, newh+8)
                    cr.fill()
                else:
                    cr.rectangle(0, offsety, w, h)
                    cr.fill()

                self.window.shape_combine_mask(bitmap, 0, 0)
            else:
                newh= (self.HIGH*self.ICONSIZE)
                bitmap = gtk.gdk.Pixmap(None, w, h, 1)
                cr = bitmap.cairo_create()

                # Clear the bitmap to False
                cr.set_source_rgb(0, 0, 0)
                cr.set_operator(cairo.OPERATOR_DEST_OUT)
                cr.paint()

                # Draw our shape into the bitmap using cairo
                cr.set_operator(cairo.OPERATOR_OVER)
                cr.rectangle(0, offsety, w, newh)
                cr.fill()
                self.window.shape_combine_mask(bitmap, 0, 0)

                self.window.merge_child_shapes()

        return returnvar

    def tr__sendEvent(self, win, ctype, data, mask=None):
    #------------------------------------------------
        """ Send a ClientMessage event to the root """
        data = (data+[0]*(5-len(data)))[:5]
        ev = Xlib.protocol.event.ClientMessage(window=win,
                       client_type=ctype, data=(32, (data)))

        if not mask:
            mask = (X.SubstructureRedirectMask|X.SubstructureNotifyMask)
        self.root.send_event(ev, event_mask=mask)

    def tr__testTiming(self,var,var2):
        # Event "loop"
        # called every 1/10th second, does all events and quits
        # quickest hack towards multi-threading i had ;)
        while self.dsp.pending_events()>0:
            e = self.dsp.next_event()
            if e.type == X.ButtonRelease:
                if(e.detail == 3):
                    # Button 3 is right click.
                    # I probably shouldnt have hard coded "3" into it...
                    # chances are I will regret this later :D
                    self.dockmenu.show_all()
                    self.dockmenu.popup(None, None, None, 3, e.time)
            if e.type == X.DestroyNotify:
                if self.tr__taskDelete(e.window.id):
                    self.tr__updatePanel(self.root, self.wind)
            if e.type == X.ConfigureNotify:
                task = self.tray.tasks[e.window.id]
                task.obj.configure(onerror=self.error,
                     width=self.ICONSIZE, height=self.ICONSIZE)
                self.tr__updatePanel(self.root, self.wind)
            if e.type == X.Expose and e.count==0:
                if(e.window.id==self.wind.id):
                    #self.wind.clear_area(0, 0, 0, 0)
                    self.tr__updatePanel(self.root, self.wind)
            if e.type == X.ClientMessage:
                data = e.data[1][1]
                task = e.data[1][2]
                if e.client_type == self._OPCODE and data == 0:
                    obj = self.dsp.create_resource_object("window", task)
                    pid=0
                    try:
                        pidob= obj.get_property(self._PIDTHING,
                            X.AnyPropertyType, 0, 1024)
                        pid = pidob.value[0]
                    except:
                        pass
                    # we either get its Process ID, or an X-error
                    # Yay :D

                    if(self.ZEROPID==False or (self.ZEROPID==True and pid>0)):

                        obj.reparent(self.tray.window.id, 0, 0)
                        ourmask = (X.ExposureMask|X.StructureNotifyMask)
                        obj.change_attributes(event_mask=ourmask)
                        self.tray.tasks[task] = Obj(obj=obj, x=0, y=0,
                            width=0, height=self.ICONSIZE, pid=pid)
                        self.tray.order.append(task)
                        self.tr__updatePanel(self.root, self.wind)
        return True

    def tr__setProps(self, dsp, win):
    #----------------------------
        """ Set necessary X atoms and panel window properties """
        self._ABOVE = dsp.intern_atom("_NET_WM_STATE_ABOVE")
        self._BELOW = dsp.intern_atom("_NET_WM_STATE_BELOW")
        self._BLACKBOX = dsp.intern_atom("_BLACKBOX_ATTRIBUTES")
        self._CHANGE_STATE = dsp.intern_atom("WM_CHANGE_STATE")
        self._CLIENT_LIST = dsp.intern_atom("_NET_CLIENT_LIST")
        self._CURRENT_DESKTOP = dsp.intern_atom("_NET_CURRENT_DESKTOP")
        self._DESKTOP = dsp.intern_atom("_NET_WM_DESKTOP")
        self._DESKTOP_COUNT = dsp.intern_atom("_NET_NUMBER_OF_DESKTOPS")
        self._DESKTOP_NAMES = dsp.intern_atom("_NET_DESKTOP_NAMES")
        self._HIDDEN = dsp.intern_atom("_NET_WM_STATE_HIDDEN")
        self._ICON = dsp.intern_atom("_NET_WM_ICON")
        self._NAME = dsp.intern_atom("_NET_WM_NAME")
        self._RPM = dsp.intern_atom("_XROOTPMAP_ID")
        self._SHADED = dsp.intern_atom("_NET_WM_STATE_SHADED")
        self._SHOWING_DESKTOP = dsp.intern_atom("_NET_SHOWING_DESKTOP")
        self._SKIP_PAGER = dsp.intern_atom("_NET_WM_STATE_SKIP_PAGER")
        self._SKIP_TASKBAR = dsp.intern_atom("_NET_WM_STATE_SKIP_TASKBAR")
        self._STATE = dsp.intern_atom("_NET_WM_STATE")
        self._STICKY = dsp.intern_atom("_NET_WM_STATE_STICKY")
        self._STRUT = dsp.intern_atom("_NET_WM_STRUT")
        self._STRUTP = dsp.intern_atom("_NET_WM_STRUT_PARTIAL")
        self._WMSTATE = dsp.intern_atom("WM_STATE")
        self._PIDTHING = dsp.intern_atom("_NET_WM_PID")

        win.set_wm_name("PyNot")
        win.set_wm_class("PyNot", "PyNot")
        win.set_wm_hints(flags=(Xutil.InputHint|Xutil.StateHint),
            input=0, initial_state=1)
        win.set_wm_normal_hints(flags=(
            Xutil.PPosition|Xutil.PMaxSize|Xutil.PMinSize),
            min_width=80, min_height=48,
            max_width=2000, max_height=48)
        win.change_property(dsp.intern_atom("_WIN_STATE"),
            Xatom.CARDINAL, 32, [1])
        win.change_property(dsp.intern_atom("_MOTIF_WM_HINTS"),
            dsp.intern_atom("_MOTIF_WM_HINTS"), 32, [0x2, 0x0, 0x0, 0x0, 0x0])
        win.change_property(self._DESKTOP, Xatom.CARDINAL, 32, [0xffffffffL])
        win.change_property(dsp.intern_atom("_NET_WM_WINDOW_TYPE"),
            Xatom.ATOM, 32, [dsp.intern_atom("_NET_WM_WINDOW_TYPE_UTILITY")])

    def cleanup(self):
        # This is my attempt to cleanly close, in such a way that the icons do
        # not get an X window error

        for tid in self.tray.order:
            t = self.tray.tasks[tid]
            g= t.obj.query_tree()
            t.obj.unmap()
            t.obj.unmap_sub_windows()
            t.obj.reparent(g.root.id, 0, 0)
        self.dsp.flush()
        return None


    def redraw(self):
        if self.USEGTK == 0 and self.USEIM == 0:
            self.modify_bg(gtk.STATE_NORMAL,
                           gtk.gdk.color_parse("#"+self.BG_COLOR[2:8]))
        elif self.USEGTK == 1:
            self.modify_bg(gtk.STATE_NORMAL,None)
        else: #USE_IM == 1
            image = None
            if IMPATH in [None,'']:
                image=gdk.pixbuf_new_from_file(D_IMPATH)
            else:
                image=gdk.pixbuf_new_from_file(IMPATH)
            (pic, mask) = image.render_pixmap_and_mask()
            self.window.set_back_pixmap(pic, False)
        self.window.clear_area_e(0, 0,
                            self.allocation.width, self.allocation.height)



gobject.type_register(mywidget)
# Register it as a widget
